"""Tests for the standard set of rules."""

import os
import os.path
from pathlib import Path

import pytest

from sqlfluff.core import Linter
from sqlfluff.core.config import FluffConfig
from sqlfluff.utils.testing.rules import assert_rule_raises_violations_in_file


@pytest.mark.parametrize(
    "rule,path,violations",
    [
        # Group By
        ("AM01", "models/my_new_project/select_distinct_group_by.sql", [(1, 8)]),
        # Multiple trailing newline
        ("LT12", "models/my_new_project/multiple_trailing_newline.sql", [(3, 1)]),
    ],
)
def test__rules__std_file_dbt(
    rule, path, violations, project_dir, dbt_fluff_config
):  # noqa
    """Test linter finds the given errors in (and only in) the right places (DBT)."""
    assert_rule_raises_violations_in_file(
        rule=rule,
        fpath=os.path.join(project_dir, path),
        violations=violations,
        fluff_config=FluffConfig(configs=dbt_fluff_config, overrides=dict(rules=rule)),
    )


def test__rules__fix_utf8(project_dir, dbt_fluff_config):  # noqa
    """Verify that non-ASCII characters are preserved by 'fix'."""
    rule = "CP01"
    path = "models/my_new_project/utf8/test.sql"
    linter = Linter(
        config=FluffConfig(configs=dbt_fluff_config, overrides=dict(rules=rule))
    )
    result = linter.lint_path(os.path.join(project_dir, path), fix=True)
    # Check that we did actually find issues.
    # NOTE: This test is mostly useful to distinguish between whether there's
    # a problem with the rule - or a problem with the file.
    record_map = {record["filepath"]: record for record in result.as_records()}
    print("Result Map: ", record_map)
    qual_path = os.path.normpath(Path(project_dir) / path)
    assert qual_path in record_map, f"{path} not in result."
    assert record_map[qual_path]["violations"], f"No issues found for {qual_path}."
    result.persist_changes(fixed_file_suffix="FIXED")
    # TODO: Check contents of file:
    # ./plugins/sqlfluff-templater-dbt/test/fixtures/dbt/dbt_project/models/
    # my_new_project/utf8/testFIXED.sql
    # Against a git file, similar to the autofix tests
    fixed_path = Path(project_dir) / "models/my_new_project/utf8/testFIXED.sql"
    cmp_filepath = Path(project_dir) / "models/my_new_project/utf8/test.sql.fixed"
    fixed_buff = fixed_path.read_text("utf8")
    comp_buff = cmp_filepath.read_text("utf8")

    # Assert that we fixed as expected
    assert fixed_buff == comp_buff
    os.unlink(fixed_path)


def test__rules__order_by(project_dir, dbt_fluff_config):  # noqa
    """Verify that rule AM03 works with dbt."""
    rule = "AM03"
    path = "models/my_new_project/AM03_test.sql"
    lntr = Linter(
        config=FluffConfig(configs=dbt_fluff_config, overrides=dict(rules=rule))
    )
    lnt = lntr.lint_path(os.path.join(project_dir, path))

    violations = lnt.check_tuples()
    assert len(violations) == 0


def test__rules__indent_oscillate(project_dir, dbt_fluff_config):  # noqa
    """Verify that we don't get oscillations with LT02 and dbt."""
    # This *should* be the wrong format
    path_1 = "models/my_new_project/indent_loop_4.sql"
    # This *should* be the correct format
    path_2 = "models/my_new_project/indent_loop_8.sql"
    # Get the content of the latter
    with open(os.path.join(project_dir, path_2), "r") as f:
        path_2_content = f.read()
    linter = Linter(
        config=FluffConfig(configs=dbt_fluff_config, overrides={"rules": "LT02"})
    )
    # Check the wrong one first (path_1)
    linted_dir = linter.lint_path(os.path.join(project_dir, path_1), fix=True)
    linted_file = linted_dir.files[0]
    assert linted_file.check_tuples() == [("LT02", 6, 1)]
    fixed_file_1, _ = linted_file.fix_string()
    assert (
        fixed_file_1 == path_2_content
    ), "indent_loop_4.sql should match indent_loop_8.sql post fix"
    # Check the correct one second, we shouldn't get any issues.
    # NOTE: This also checks that the fixed version of the first one wouldn't
    # change again.
    linted_dir = linter.lint_path(os.path.join(project_dir, path_2), fix=True)
    linted_file = linted_dir.files[0]
    assert linted_file.check_tuples() == []  # Should find no issues.



"""MariaDB Dialect.

https://mariadb.com/kb/en/sql-statements-structure/
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    AnySetOf,
    BaseSegment,
    Bracketed,
    Dedent,
    Delimited,
    Indent,
    Matchable,
    OneOf,
    OptionallyBracketed,
    ParseMode,
    Ref,
    Sequence,
)
from sqlfluff.dialects import dialect_mysql as mysql
from sqlfluff.dialects.dialect_mariadb_keywords import (
    mariadb_reserved_keywords,
    mariadb_unreserved_keywords,
)

# ansi_dialect = load_raw_dialect("ansi")
mysql_dialect = load_raw_dialect("mysql")
mariadb_dialect = mysql_dialect.copy_as(
    "mariadb",
    formatted_name="MariaDB",
    docstring="""**Default Casing**: ``lowercase``

**Quotes**: String Literals: ``''``, ``""`` or ``@``,
Identifiers: |back_quotes|.

The dialect for `MariaDB <https://www.mariadb.org/>`_.""",
)
mariadb_dialect.update_keywords_set_from_multiline_string(
    "unreserved_keywords", mariadb_unreserved_keywords
)
mariadb_dialect.sets("reserved_keywords").clear()
mariadb_dialect.update_keywords_set_from_multiline_string(
    "reserved_keywords", mariadb_reserved_keywords
)


class ColumnConstraintSegment(mysql.ColumnConstraintSegment):
    """A column option; each CREATE TABLE column can have 0 or more."""

    match_grammar: Matchable = OneOf(
        mysql.ColumnConstraintSegment.match_grammar,
        Sequence(
            Sequence("GENERATED", "ALWAYS", optional=True),
            "AS",
            Bracketed(Ref("ExpressionSegment")),
            OneOf("PERSISTENT", "STORED", "VIRTUAL", optional=True),
        ),
    )


class CreateUserStatementSegment(mysql.CreateUserStatementSegment):
    """`CREATE USER` statement.

    https://mariadb.com/kb/en/create-user/
    """

    match_grammar = mysql.CreateUserStatementSegment.match_grammar.copy(
        insert=[Ref("OrReplaceGrammar", optional=True)],
        before=Ref.keyword("USER"),
    )


class CreateTableStatementSegment(mysql.CreateTableStatementSegment):
    """`CREATE TABLE` segment.

    https://mariadb.com/kb/en/create-table/
    """

    ES = Ref("ExpressionSegment")
    CRS = Ref("ColumnReferenceSegment")
    NLS = Ref("NumericLiteralSegment")
    ORS = Ref("ObjectReferenceSegment")
    TRS = Ref("TableReferenceSegment")
    SQIS = Ref("SingleQuotedIdentifierSegment")

    match_grammar = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        Ref("TemporaryTransientGrammar", optional=True),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        OneOf(
            # Columns and comment syntax:
            Sequence(
                Bracketed(
                    Delimited(
                        OneOf(
                            Ref("TableConstraintSegment"),
                            Ref("ColumnDefinitionSegment"),
                        ),
                    )
                ),
                Ref("CommentClauseSegment", optional=True),
                Sequence(
                    Ref.keyword("AS", optional=True),
                    OptionallyBracketed(Ref("SelectableGrammar")),
                    optional=True,
                ),
            ),
            # Create AS syntax:
            Sequence(
                Ref.keyword("AS", optional=True),
                OptionallyBracketed(Ref("SelectableGrammar")),
            ),
            # Create like syntax
            Sequence("LIKE", Ref("TableReferenceSegment")),
        ),
        Ref("TableEndClauseSegment", optional=True),
        AnyNumberOf(
            Sequence(
                Ref.keyword("DEFAULT", optional=True),
                OneOf(
                    Ref("ParameterNameSegment"),
                    Sequence("CHARACTER", "SET"),
                    Sequence(OneOf("DATA", "INDEX"), "DIRECTORY"),
                    Sequence("WITH", "SYSTEM"),
                ),
                Ref("EqualsSegment", optional=True),
                OneOf(
                    Ref("LiteralGrammar"),
                    Ref("ParameterNameSegment"),
                    Ref("QuotedLiteralSegment"),
                    Ref("SingleQuotedIdentifierSegment"),
                    Ref("NumericLiteralSegment"),
                    # Union option
                    Bracketed(
                        Delimited(Ref("TableReferenceSegment")),
                    ),
                ),
            ),
            # Partition Options
            # https://dev.mysql.com/doc/refman/8.0/en/create-table.html#create-table-partitioning
            Sequence(
                "PARTITION",
                "BY",
                OneOf(
                    Sequence(
                        Ref.keyword("LINEAR", optional=True),
                        OneOf(
                            Sequence("HASH", Ref("ExpressionSegment")),
                            Sequence(
                                "KEY",
                                Sequence(
                                    "ALGORITHM",
                                    Ref("EqualsSegment"),
                                    Ref("NumericLiteralSegment"),
                                    optional=True,
                                ),
                                Delimited(Ref("ColumnReferenceSegment")),
                            ),
                        ),
                    ),
                    Sequence(
                        OneOf("RANGE", "LIST"),
                        OneOf(
                            Ref("ExpressionSegment"),
                            Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                        ),
                    ),
                ),
                Sequence("PARTITIONS", Ref("NumericLiteralSegment"), optional=True),
                Sequence(
                    "SUBPARTITION",
                    "BY",
                    Sequence(
                        Ref.keyword("LINEAR", optional=True),
                        OneOf(
                            Sequence("HASH", Ref("ExpressionSegment")),
                            Sequence(
                                "KEY",
                                Sequence(
                                    "ALGORITHM",
                                    Ref("EqualsSegment"),
                                    Ref("NumericLiteralSegment"),
                                    optional=True,
                                ),
                                Bracketed(Ref("ColumnReferenceSegment")),
                            ),
                        ),
                    ),
                    Sequence(
                        "SUBPARTITIONS",
                        Ref("NumericLiteralSegment"),
                        optional=True,
                    ),
                    optional=True,
                ),
                # optional partition_definition(s)
                AnyNumberOf(
                    Bracketed(
                        Delimited(
                            Sequence(
                                "PARTITION",
                                Ref("ColumnReferenceSegment"),
                                AnyNumberOf(
                                    Sequence(
                                        "VALUES",
                                        OneOf(
                                            Sequence(
                                                "LESS",
                                                "THAN",
                                                OneOf(
                                                    "MAXVALUE",
                                                    Bracketed(
                                                        OneOf(
                                                            ES,
                                                            CRS,
                                                            NLS,
                                                            Ref("LiteralGrammar"),
                                                        ),
                                                    ),
                                                ),
                                            ),
                                            Sequence(
                                                "IN",
                                                Bracketed(
                                                    Ref("ObjectReferenceSegment")
                                                ),
                                            ),
                                        ),
                                    ),
                                    Sequence(
                                        OneOf(
                                            Ref("ParameterNameSegment"),
                                            Sequence("CHARACTER", "SET"),
                                            Sequence(
                                                OneOf("DATA", "INDEX"),
                                                "DIRECTORY",
                                            ),
                                            Sequence("WITH", "SYSTEM"),
                                        ),
                                        Ref("EqualsSegment", optional=True),
                                        OneOf(
                                            Ref("LiteralGrammar"),
                                            Ref("ParameterNameSegment"),
                                            Ref("QuotedLiteralSegment"),
                                            Ref("SingleQuotedIdentifierSegment"),
                                            Ref("NumericLiteralSegment"),
                                            # Union option
                                            Bracketed(
                                                Delimited(Ref("TableReferenceSegment")),
                                            ),
                                        ),
                                    ),
                                    # optional subpartition_definition(s)
                                    Sequence(
                                        Ref.keyword("SUBPARTITION", optional=True),
                                        Ref("LiteralGrammar"),
                                        AnyNumberOf(
                                            Sequence(
                                                "VALUES",
                                                OneOf(
                                                    Sequence(
                                                        "LESS",
                                                        "THAN",
                                                        OneOf(
                                                            "MAXVALUE",
                                                            Bracketed(ES),
                                                            Bracketed(CRS),
                                                        ),
                                                    ),
                                                    Sequence(
                                                        "IN",
                                                        Bracketed(ORS),
                                                    ),
                                                ),
                                            ),
                                            Sequence(
                                                OneOf(
                                                    Ref("ParameterNameSegment"),
                                                    Sequence("CHARACTER", "SET"),
                                                    Sequence(
                                                        OneOf("DATA", "INDEX"),
                                                        "DIRECTORY",
                                                    ),
                                                    Sequence("WITH", "SYSTEM"),
                                                ),
                                                Ref(
                                                    "EqualsSegment",
                                                    optional=True,
                                                ),
                                                OneOf(
                                                    Ref("LiteralGrammar"),
                                                    Ref("ParameterNameSegment"),
                                                    Ref("QuotedLiteralSegment"),
                                                    SQIS,
                                                    Ref("NumericLiteralSegment"),
                                                    # Union option
                                                    Bracketed(
                                                        Delimited(TRS),
                                                    ),
                                                ),
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )


class DeleteStatementSegment(BaseSegment):
    """A `DELETE` statement.

    https://mariadb.com/kb/en/delete/
    """

    type = "delete_statement"
    match_grammar = Sequence(
        "DELETE",
        Ref.keyword("LOW_PRIORITY", optional=True),
        Ref.keyword("QUICK", optional=True),
        Ref.keyword("IGNORE", optional=True),
        OneOf(
            Sequence(
                "FROM",
                Delimited(
                    Ref("DeleteTargetTableSegment"),
                    terminators=["USING"],
                ),
                Ref("DeleteUsingClauseSegment"),
                Ref("WhereClauseSegment", optional=True),
            ),
            Sequence(
                Delimited(
                    Ref("DeleteTargetTableSegment"),
                    terminators=["FROM"],
                ),
                Ref("FromClauseSegment"),
                Ref("WhereClauseSegment", optional=True),
            ),
            Sequence(
                Ref("FromClauseSegment"),
                Ref("SelectPartitionClauseSegment", optional=True),
                Ref("WhereClauseSegment", optional=True),
                Ref("OrderByClauseSegment", optional=True),
                Ref("LimitClauseSegment", optional=True),
                Ref("ReturningClauseSegment", optional=True),
            ),
        ),
    )


class FlushStatementSegment(mysql.FlushStatementSegment):
    """A `Flush` statement.

    https://mariadb.com/kb/en/flush/
    """

    match_grammar: Matchable = Sequence(
        "FLUSH",
        OneOf(
            "NO_WRITE_TO_BINLOG",
            "LOCAL",
            optional=True,
        ),
        OneOf(
            Delimited(
                Sequence("BINARY", "LOGS"),
                Sequence("ENGINE", "LOGS"),
                Sequence("ERROR", "LOGS"),
                Sequence("GENERAL", "LOGS"),
                Sequence("QUERY", "CACHE"),
                Sequence("SLOW", "LOGS"),
                Sequence(Ref.keyword("RESET", optional=True), "MASTER"),
                Sequence(OneOf("GLOBAL", "SESSION", optional=True), "STATUS"),
                Sequence(
                    "RELAY",
                    "LOGS",
                    Sequence("FOR", "CHANNEL", optional=True),
                    Ref("ObjectReferenceSegment"),
                ),
                "HOSTS",
                "LOGS",
                "PRIVILEGES",
                "CHANGED_PAGE_BITMAPS",
                "CLIENT_STATISTICS",
                "DES_KEY_FILE",
                "INDEX_STATISTICS",
                "QUERY_RESPONSE_TIME",
                "SLAVE",
                "SSL",
                "TABLE_STATISTICS",
                "USER_STATISTICS",
                "USER_VARIABLES",
                "USER_RESOURCES",
            ),
            Sequence(
                "TABLES",
                Sequence(
                    Delimited(Ref("TableReferenceSegment"), terminators=["WITH"]),
                    optional=True,
                ),
                Sequence(
                    "WITH",
                    "READ",
                    "LOCK",
                    Sequence("AND", "DISABLE", "CHECKPOINT", optional=True),
                    optional=True,
                ),
            ),
            Sequence(
                "TABLES",
                Sequence(
                    Delimited(Ref("TableReferenceSegment"), terminators=["FOR"]),
                    optional=False,
                ),
                Sequence("FOR", "EXPORT", optional=True),
            ),
        ),
    )


class GroupByClauseSegment(BaseSegment):
    """A `GROUP BY` clause like in `SELECT`."""

    type = "groupby_clause"

    match_grammar: Matchable = Sequence(
        "GROUP",
        "BY",
        Indent,
        Sequence(
            Delimited(
                Sequence(
                    OneOf(
                        Ref("ColumnReferenceSegment"),
                        # Can `GROUP BY 1`
                        Ref("NumericLiteralSegment"),
                        # Can `GROUP BY coalesce(col, 1)`
                        Ref("ExpressionSegment"),
                    ),
                    OneOf("ASC", "DESC", optional=True),
                ),
                terminators=[Ref("GroupByClauseTerminatorGrammar")],
            ),
        ),
        Ref("WithRollupClauseSegment", optional=True),
        Dedent,
    )


class InsertStatementSegment(BaseSegment):
    """An `INSERT` statement.

    https://mariadb.com/kb/en/insert/
    """

    type = "insert_statement"
    match_grammar = Sequence(
        "INSERT",
        OneOf(
            "LOW_PRIORITY",
            "DELAYED",
            "HIGH_PRIORITY",
            optional=True,
        ),
        Ref.keyword("IGNORE", optional=True),
        Ref.keyword("INTO", optional=True),
        Ref("TableReferenceSegment"),
        Sequence(
            "PARTITION",
            Bracketed(
                Ref("SingleIdentifierListSegment"),
            ),
            optional=True,
        ),
        Ref("BracketedColumnReferenceListGrammar", optional=True),
        AnySetOf(
            OneOf(
                Ref("ValuesClauseSegment"),
                Ref("SetClauseListSegment"),
                Ref("SelectStatementSegment"),
                optional=False,
            ),
            Ref("InsertRowAliasSegment", optional=True),
            Ref("UpsertClauseListSegment", optional=True),
            Ref("ReturningClauseSegment", optional=True),
        ),
    )


class ReplaceSegment(BaseSegment):
    """A `REPLACE` statement.

    As per https://mariadb.com/kb/en/replace/
    """

    type = "replace_statement"

    match_grammar = Sequence(
        "REPLACE",
        OneOf("LOW_PRIORITY", "DELAYED", optional=True),
        Sequence("INTO", optional=True),
        Ref("TableReferenceSegment"),
        Ref("SelectPartitionClauseSegment", optional=True),
        OneOf(
            Sequence(
                Ref("BracketedColumnReferenceListGrammar", optional=True),
                Ref("ValuesClauseSegment"),
            ),
            Ref("SetClauseListSegment"),
            Sequence(
                Ref("BracketedColumnReferenceListGrammar", optional=True),
                Ref("SelectStatementSegment"),
            ),
        ),
        Ref("ReturningClauseSegment", optional=True),
    )


class ReturningClauseSegment(BaseSegment):
    """This is a `RETURNING` clause.

    A RETURNING clause returns values modified by a
    INSERT, DELETE or REPLACE query.

    https://mariadb.com/kb/en/insert/
    https://mariadb.com/kb/en/delete/
    https://mariadb.com/kb/en/replace/
    """

    type = "returning_clause"

    match_grammar: Matchable = Sequence(
        "RETURNING",
        Indent,
        Delimited(
            Ref("SelectClauseElementSegment"),
            allow_trailing=True,
        ),
        Dedent,
        terminators=[Ref("SelectClauseTerminatorGrammar")],
        parse_mode=ParseMode.GREEDY_ONCE_STARTED,
    )


class SelectStatementSegment(mysql.SelectStatementSegment):
    """A `SELECT` statement.

    https://mariadb.com/kb/en/select/
    """

    # Inherit most of the parse grammar from the original.
    match_grammar = mysql.UnorderedSelectStatementSegment.match_grammar.copy(
        insert=[
            Ref("OrderByClauseSegment", optional=True),
            Ref("LimitClauseSegment", optional=True),
            Ref("NamedWindowSegment", optional=True),
            Ref("IntoClauseSegment", optional=True),
        ],
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("UpsertClauseListSegment"),
            Ref("WithCheckOptionSegment"),
            Ref("ReturningClauseSegment"),
        ],
        # Overwrite the terminators, because we want to remove some from the
        # expression above.
        replace_terminators=True,
    )


class WithRollupClauseSegment(BaseSegment):
    """A `WITH ROLLUP` clause after the `GROUP BY` clause."""

    type = "with_rollup_clause"

    match_grammar = Sequence(
        "WITH",
        "ROLLUP",
    )



"""Tests for PositionMarker."""

import pytest

from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.templaters import TemplatedFile


@pytest.mark.parametrize(
    "raw,start_pos,end_pos",
    [
        ("fsaljk", (0, 0), (0, 6)),
        ("", (2, 2), (2, 2)),
        # NB: 1 indexed, not 0 indexed.
        ("\n", (2, 2), (3, 1)),
        ("boo\n", (2, 2), (3, 1)),
        ("boo\nfoo", (2, 2), (3, 4)),
        ("\nfoo", (2, 2), (3, 4)),
    ],
)
def test_markers__infer_next_position(raw, start_pos, end_pos):
    """Test that we can correctly infer positions from strings."""
    assert end_pos == PositionMarker.infer_next_position(raw, *start_pos)


def test_markers__setting_position_raw():
    """Test that we can correctly infer positions from strings & locations."""
    templ = TemplatedFile.from_string("foobar")
    # Check inference in the template
    assert templ.get_line_pos_of_char_pos(2, source=True) == (1, 3)
    assert templ.get_line_pos_of_char_pos(2, source=False) == (1, 3)
    # Now check it passes through
    pos = PositionMarker(slice(2, 5), slice(2, 5), templ)
    # Can we infer positions correctly?
    assert pos.working_loc == (1, 3)
    # Check other marker properties work too (i.e. source properties)
    assert pos.line_no == 1
    assert pos.line_pos == 3  # i.e. 2 + 1 (for 1-indexed)


def test_markers__setting_position_working():
    """Test that we can correctly set positions manually."""
    templ = TemplatedFile.from_string("foobar")
    pos = PositionMarker(slice(2, 5), slice(2, 5), templ, 4, 4)
    # Can we don't infer when we're explicitly told.
    assert pos.working_loc == (4, 4)


def test_markers__comparison():
    """Test that we can correctly compare markers."""
    templ = TemplatedFile.from_string("abc")
    # Make position markers for each of a, b & c
    # NOTE: We're not explicitly setting the working location, we
    # rely here on the marker inferring that correctly itself.
    a_pos = PositionMarker(slice(0, 1), slice(0, 1), templ)
    b_pos = PositionMarker(slice(1, 2), slice(1, 2), templ)
    c_pos = PositionMarker(slice(2, 3), slice(2, 3), templ)
    all_pos = (a_pos, b_pos, c_pos)
    # Check equality
    assert all(p == p for p in all_pos)
    # Check inequality
    assert a_pos != b_pos and a_pos != c_pos and b_pos != c_pos
    # Check less than
    assert a_pos < b_pos and b_pos < c_pos
    assert not c_pos < a_pos
    # Check greater than
    assert c_pos > a_pos and c_pos > b_pos
    assert not a_pos > c_pos
    # Check less than or equal
    assert all(a_pos <= p for p in all_pos)
    # Check greater than or equal
    assert all(c_pos >= p for p in all_pos)



"""Test routines for fixing errors."""

import logging

import pytest

from sqlfluff.core.linter.fix import compute_anchor_edit_info
from sqlfluff.core.linter.patch import FixPatch, generate_source_patches
from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.parser.segments import (
    BaseSegment,
    RawSegment,
    TemplateSegment,
)
from sqlfluff.core.parser.segments.raw import SourceFix
from sqlfluff.core.rules.fix import LintFix
from sqlfluff.core.templaters import RawFileSlice, TemplatedFile
from sqlfluff.core.templaters.base import TemplatedFileSlice


@pytest.fixture(scope="module")
def raw_segments(generate_test_segments):
    """Construct a list of raw segments as a fixture."""
    return generate_test_segments(["foobar", ".barfoo"])


def test__rules_base_segments_compute_anchor_edit_info(raw_segments):
    """Test BaseSegment.compute_anchor_edit_info()."""
    # Construct a fix buffer, intentionally with:
    # - one duplicate.
    # - two different incompatible fixes on the same segment.
    fixes = [
        LintFix.replace(raw_segments[0], [raw_segments[0].edit(raw="a")]),
        LintFix.replace(raw_segments[0], [raw_segments[0].edit(raw="a")]),
        LintFix.replace(raw_segments[0], [raw_segments[0].edit(raw="b")]),
    ]
    anchor_info_dict = compute_anchor_edit_info(fixes)
    # Check the target segment is the only key we have.
    assert list(anchor_info_dict.keys()) == [raw_segments[0].uuid]
    anchor_info = anchor_info_dict[raw_segments[0].uuid]
    # Check that the duplicate as been deduplicated.
    # i.e. this isn't 3.
    assert anchor_info.replace == 2
    # Check the fixes themselves.
    # NOTE: There's no duplicated first fix.
    assert anchor_info.fixes == [
        LintFix.replace(raw_segments[0], [raw_segments[0].edit(raw="a")]),
        LintFix.replace(raw_segments[0], [raw_segments[0].edit(raw="b")]),
    ]
    # Check the first replace
    assert anchor_info._first_replace == LintFix.replace(
        raw_segments[0], [raw_segments[0].edit(raw="a")]
    )


templated_file_1 = TemplatedFile.from_string("abc")
templated_file_2 = TemplatedFile(
    "{# blah #}{{ foo }}bc",
    "<testing>",
    "abc",
    [
        TemplatedFileSlice("comment", slice(0, 10), slice(0, 0)),
        TemplatedFileSlice("templated", slice(10, 19), slice(0, 1)),
        TemplatedFileSlice("literal", slice(19, 21), slice(1, 3)),
    ],
    [
        RawFileSlice("{# blah #}", "comment", 0),
        RawFileSlice("{{ foo }}", "templated", 10),
        RawFileSlice("bc", "literal", 19),
    ],
)


@pytest.mark.parametrize(
    "tree,templated_file,expected_result",
    [
        # Trivial example
        (
            RawSegment(
                "abc",
                PositionMarker(slice(0, 3), slice(0, 3), templated_file_1),
                "code",
            ),
            templated_file_1,
            [],
        ),
        # Simple literal edit example
        (
            RawSegment(
                "abz",
                PositionMarker(slice(0, 3), slice(0, 3), templated_file_1),
                "code",
            ),
            templated_file_1,
            [FixPatch(slice(0, 3), "abz", "literal", slice(0, 3), "abc", "abc")],
        ),
        # Nested literal edit example
        (
            BaseSegment(
                [
                    RawSegment(
                        "a",
                        PositionMarker(slice(0, 1), slice(0, 1), templated_file_1),
                        "code",
                    ),
                    RawSegment(
                        "b",
                        PositionMarker(slice(1, 2), slice(1, 2), templated_file_1),
                        "code",
                    ),
                    RawSegment(
                        "z",
                        PositionMarker(slice(2, 3), slice(2, 3), templated_file_1),
                        "code",
                    ),
                ]
            ),
            templated_file_1,
            [FixPatch(slice(0, 3), "abz", "literal", slice(0, 3), "abc", "abc")],
        ),
        # More complicated templating example
        (
            BaseSegment(
                [
                    TemplateSegment(
                        PositionMarker(slice(0, 10), slice(0, 0), templated_file_2),
                        "{# blah #}",
                        "comment",
                    ),
                    RawSegment(
                        "a",
                        PositionMarker(slice(10, 20), slice(0, 1), templated_file_2),
                        "code",
                    ),
                    RawSegment(
                        "b",
                        PositionMarker(slice(19, 20), slice(1, 2), templated_file_2),
                        "code",
                    ),
                    RawSegment(
                        "z",
                        PositionMarker(slice(20, 21), slice(2, 3), templated_file_2),
                        "code",
                    ),
                ]
            ),
            templated_file_2,
            [FixPatch(slice(2, 3), "z", "literal", slice(20, 21), "c", "c")],
        ),
        # Templating example with fixes
        (
            BaseSegment(
                [
                    TemplateSegment(
                        PositionMarker(slice(0, 10), slice(0, 0), templated_file_2),
                        "{# blah #}",
                        "comment",
                        source_fixes=[
                            SourceFix("{# fixed #}", slice(0, 10), slice(0, 0))
                        ],
                    ),
                    RawSegment(
                        "a",
                        PositionMarker(slice(10, 19), slice(0, 1), templated_file_2),
                        "code",
                        source_fixes=[
                            SourceFix("{{ bar }}", slice(10, 19), slice(0, 1))
                        ],
                    ),
                    RawSegment(
                        "b",
                        PositionMarker(slice(19, 20), slice(1, 2), templated_file_2),
                        "code",
                    ),
                    RawSegment(
                        "z",
                        PositionMarker(slice(20, 21), slice(2, 3), templated_file_2),
                        "code",
                    ),
                ]
            ),
            templated_file_2,
            [
                FixPatch(
                    slice(0, 0), "{# fixed #}", "source", slice(0, 10), "", "{# blah #}"
                ),
                FixPatch(
                    slice(0, 1), "{{ bar }}", "source", slice(10, 19), "a", "{{ foo }}"
                ),
                FixPatch(slice(2, 3), "z", "literal", slice(20, 21), "c", "c"),
            ],
        ),
    ],
)
def test__fix__generate_source_patches(tree, templated_file, expected_result, caplog):
    """Test generate_source_patches.

    This is part of fix_string().
    """
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.linter"):
        result = generate_source_patches(tree, templated_file)
    assert result == expected_result



"""This is a modified log capture mechanism which reliably works.

So that logs are handled appropriately by the CLI, sqlfluff
modifies the root logger in a way that can conflict with pytest.

See: https://github.com/pytest-dev/pytest/issues/3697

This fixture returns a context manager to handle them better
and enable testing of logs while working around the restrictions
of setting the `propagate` attribute of the logger in each test.

Code adapted from:
https://github.com/pytest-dev/pytest/issues/3697#issuecomment-792129636

"""

import logging
from contextlib import contextmanager
from typing import Iterator

from _pytest.logging import LogCaptureHandler, _remove_ansi_escape_sequences


class FluffLogHandler(LogCaptureHandler):
    """A modified LogCaptureHandler which also exposes some helper functions.

    The aim is to mimic some of the methods available on caplog.

    See:
    https://docs.pytest.org/en/7.1.x/_modules/_pytest/logging.html
    """

    @property
    def text(self) -> str:
        """The formatted log text."""
        return _remove_ansi_escape_sequences(self.stream.getvalue())


@contextmanager
def fluff_log_catcher(level: int, logger_name: str) -> Iterator[FluffLogHandler]:
    """Context manager that sets the level for capturing of logs.

    After the end of the 'with' statement the level is restored to its
    original value.

    Args:
        level (int): The lowest logging level to capture.
        logger_name (str): The name of the logger to capture.
    """
    assert logger_name.startswith(
        "sqlfluff"
    ), "This should only be used with a SQLFluff logger."
    logger = logging.getLogger(logger_name)
    handler = FluffLogHandler()
    orig_level = logger.level
    logger.setLevel(level)
    logger.addHandler(handler)
    try:
        yield handler
    finally:
        logger.setLevel(orig_level)
        logger.removeHandler(handler)



"""Test the behaviour of the unparsable routines."""

from typing import Any, Optional

import pytest

from sqlfluff.core import FluffConfig
from sqlfluff.core.parser import BaseSegment, Lexer, RawSegment
from sqlfluff.core.parser.context import ParseContext


# NOTE: Being specific on the segment ref helps to avoid crazy nesting.
@pytest.mark.parametrize(
    "segmentref,dialect,raw,structure",
    [
        (
            # The first here makes sure all of this works from the outer
            # segment, but for other tests we should aim to be more specific.
            None,
            "ansi",
            "SELECT 1 1",
            (
                "file",
                (
                    (
                        "statement",
                        (
                            (
                                "select_statement",
                                (
                                    (
                                        "select_clause",
                                        (
                                            ("keyword", "SELECT"),
                                            ("whitespace", " "),
                                            (
                                                "select_clause_element",
                                                (("numeric_literal", "1"),),
                                            ),
                                            ("whitespace", " "),
                                            (
                                                "unparsable",
                                                (("numeric_literal", "1"),),
                                            ),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ),
        (
            "SelectClauseSegment",
            "ansi",
            "SELECT 1 1",
            (
                "select_clause",
                (
                    ("keyword", "SELECT"),
                    ("whitespace", " "),
                    (
                        "select_clause_element",
                        (("numeric_literal", "1"),),
                    ),
                    ("whitespace", " "),
                    # We should get a single unparsable section
                    # here at the end.
                    (
                        "unparsable",
                        (("numeric_literal", "1"),),
                    ),
                ),
            ),
        ),
        # This more complex example looks a little strange, but does
        # reflect current unparsable behaviour. During future work
        # on the parser, the structure of this result may change
        # but it should still result in am unparsable section _within_
        # the brackets, and not just a totally unparsable statement.
        (
            "SelectClauseSegment",
            "ansi",
            "SELECT 1 + (2 2 2)",
            (
                "select_clause",
                (
                    ("keyword", "SELECT"),
                    ("whitespace", " "),
                    (
                        "select_clause_element",
                        (
                            (
                                "expression",
                                (
                                    ("numeric_literal", "1"),
                                    ("whitespace", " "),
                                    ("binary_operator", "+"),
                                    ("whitespace", " "),
                                    (
                                        "bracketed",
                                        (
                                            ("start_bracket", "("),
                                            ("expression", (("numeric_literal", "2"),)),
                                            ("whitespace", " "),
                                            (
                                                "unparsable",
                                                (
                                                    ("numeric_literal", "2"),
                                                    ("whitespace", " "),
                                                    ("numeric_literal", "2"),
                                                ),
                                            ),
                                            ("end_bracket", ")"),
                                        ),
                                    ),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    ],
)
def test_dialect_unparsable(
    segmentref: Optional[str], dialect: str, raw: str, structure: Any
):
    """Test the structure of unparsables."""
    config = FluffConfig(overrides=dict(dialect=dialect))

    # Get the referenced object (if set, otherwise root)
    if segmentref:
        Seg = config.get("dialect_obj").ref(segmentref)
    else:
        Seg = config.get("dialect_obj").get_root_segment()
    # We only allow BaseSegments as matchables in this test.
    assert issubclass(Seg, BaseSegment)
    assert not issubclass(Seg, RawSegment)

    # Lex the raw string.
    lex = Lexer(config=config)
    segments, vs = lex.lex(raw)
    assert not vs
    # Strip the end of file token if it's there. It will
    # confuse most segments.
    if segmentref and segments[-1].is_type("end_of_file"):
        segments = segments[:-1]

    ctx = ParseContext.from_config(config)

    # Match against the segment.
    match = Seg.match(segments, 0, ctx)
    result = match.apply(segments)

    assert len(result) == 1
    parsed = result[0]
    assert isinstance(parsed, Seg)

    assert parsed.to_tuple(show_raw=True) == structure



"""Raw segment definitions.

This is designed to be the root segment, without
any children, and the output of the lexer.
"""

from typing import Any, Callable, Dict, FrozenSet, List, Optional, Tuple, Union, cast
from uuid import uuid4

import regex as re

from sqlfluff.core.parser.markers import PositionMarker
from sqlfluff.core.parser.segments.base import BaseSegment, SourceFix


class RawSegment(BaseSegment):
    """This is a segment without any subsegments."""

    type = "raw"
    _is_code = True
    _is_comment = False
    _is_whitespace = False
    # Classes inheriting from RawSegment may provide a _default_raw
    # to enable simple initialisation.
    _default_raw = ""

    def __init__(
        self,
        raw: Optional[str] = None,
        pos_marker: Optional[PositionMarker] = None,
        # For legacy and syntactic sugar we allow the simple
        # `type` argument here, but for more precise inheritance
        # we suggest using the `instance_types` option.
        type: Optional[str] = None,
        instance_types: Tuple[str, ...] = (),
        trim_start: Optional[Tuple[str, ...]] = None,
        trim_chars: Optional[Tuple[str, ...]] = None,
        source_fixes: Optional[List[SourceFix]] = None,
        uuid: Optional[int] = None,
        quoted_value: Optional[Tuple[str, Union[int, str]]] = None,
        escape_replacements: Optional[List[Tuple[str, str]]] = None,
        casefold: Optional[Callable[[str], str]] = None,
    ):
        """Initialise raw segment.

        If raw is not provided, we default to _default_raw if present.
        If pos_marker is not provided, it is assume that this will be
        inserted later as part of a reposition phase.
        """
        if raw is not None:  # NB, raw *can* be an empty string and be valid
            self._raw = raw
        else:
            self._raw = self._default_raw
        self._raw_upper = self._raw.upper()
        # pos marker is required here. We ignore the typing initially
        # because it might *initially* be unset, but it will be reset
        # later.
        self.pos_marker: PositionMarker = pos_marker  # type: ignore
        # Set the segments attribute to be an empty tuple.
        self.segments = ()
        self.instance_types: Tuple[str, ...]
        if type:
            assert not instance_types, "Cannot set `type` and `instance_types`."
            self.instance_types = (type,)
        else:
            self.instance_types = instance_types
        # What should we trim off the ends to get to content
        self.trim_start = trim_start
        self.trim_chars = trim_chars
        # Keep track of any source fixes
        self._source_fixes = source_fixes
        # UUID for matching (the int attribute of it)
        self.uuid = uuid or uuid4().int
        self.representation = "<{}: ({}) {!r}>".format(
            self.__class__.__name__, self.pos_marker, self.raw
        )
        self.quoted_value = quoted_value
        self.escape_replacements = escape_replacements
        self.casefold = casefold
        self._raw_value: str = self._raw_normalized()

    def __repr__(self) -> str:
        # This is calculated at __init__, because all elements are immutable
        # and this was previously recalculating the pos marker,
        # and became very expensive
        return self.representation

    def __setattr__(self, key: str, value: Any) -> None:
        """Overwrite BaseSegment's __setattr__ with BaseSegment's superclass."""
        super(BaseSegment, self).__setattr__(key, value)

    # ################ PUBLIC PROPERTIES

    @property
    def is_code(self) -> bool:
        """Return True if this segment is code."""
        return self._is_code

    @property
    def is_comment(self) -> bool:
        """Return True if this segment is a comment."""
        return self._is_comment

    @property
    def is_whitespace(self) -> bool:
        """Return True if this segment is whitespace."""
        return self._is_whitespace

    @property
    def raw(self) -> str:
        """Returns the raw segment."""
        return self._raw

    @property
    def raw_upper(self) -> str:
        """Returns the raw segment in uppercase."""
        return self._raw_upper

    @property
    def raw_segments(self) -> List["RawSegment"]:
        """Returns self to be compatible with calls to its superclass."""
        return [self]

    @property
    def class_types(self) -> FrozenSet[str]:
        """The set of full types for this segment, including inherited.

        Add the surrogate type for raw segments.
        """
        return frozenset(self.instance_types) | super().class_types

    @property
    def source_fixes(self) -> List[SourceFix]:
        """Return any source fixes as list."""
        return self._source_fixes or []

    # ################ INSTANCE METHODS

    def invalidate_caches(self) -> None:
        """Overwrite superclass functionality."""
        pass

    def get_type(self) -> str:
        """Returns the type of this segment as a string."""
        if self.instance_types:
            return self.instance_types[0]
        return super().get_type()

    def is_type(self, *seg_type: str) -> bool:
        """Extend the parent class method with the surrogate types."""
        if set(self.instance_types).intersection(seg_type):
            return True
        return self.class_is_type(*seg_type)

    def get_raw_segments(self) -> List["RawSegment"]:
        """Iterate raw segments, mostly for searching."""
        return [self]

    def raw_trimmed(self) -> str:
        """Return a trimmed version of the raw content.

        Returns:
            str: The trimmed version of the raw content.
        """
        raw_buff = self.raw
        if self.trim_start:
            for seq in self.trim_start:
                if raw_buff.startswith(seq):
                    raw_buff = raw_buff[len(seq) :]
        if self.trim_chars:
            raw_buff = self.raw
            # for each thing to trim
            for seq in self.trim_chars:
                # trim start
                while raw_buff.startswith(seq):
                    raw_buff = raw_buff[len(seq) :]
                # trim end
                while raw_buff.endswith(seq):
                    raw_buff = raw_buff[: -len(seq)]
            return raw_buff
        return raw_buff

    def _raw_normalized(self) -> str:
        """Returns the string of the raw content's value.

        E.g. This removes leading and trailing quote characters, removes escapes

        Return:
        str: The raw content's value
        """
        raw_buff = self.raw
        if self.quoted_value:
            _match = re.match(self.quoted_value[0], raw_buff)
            if _match:
                _group_match = _match.group(self.quoted_value[1])
                if isinstance(_group_match, str):
                    raw_buff = _group_match
        if self.escape_replacements:
            for old, new in self.escape_replacements:
                raw_buff = re.sub(old, new, raw_buff)
        return raw_buff

    def raw_normalized(self, casefold: bool = True) -> str:
        """Returns a normalized string of the raw content.

        E.g. This removes leading and trailing quote characters, removes escapes,
        optionally casefolds to the dialect's casing

        Return:
        str: The normalized version of the raw content
        """
        raw_buff = self._raw_value
        if self.casefold and casefold:
            raw_buff = self.casefold(raw_buff)
        return raw_buff

    def stringify(
        self, ident: int = 0, tabsize: int = 4, code_only: bool = False
    ) -> str:
        """Use indentation to render this segment and its children as a string.

        Args:
            ident (int, optional): The indentation level. Defaults to 0.
            tabsize (int, optional): The size of each tab. Defaults to 4.
            code_only (bool, optional): Whether to render only the code.
                Defaults to False.

        Returns:
            str: The rendered string.
        """
        preface = self._preface(ident=ident, tabsize=tabsize)
        return preface + "\n"

    def _suffix(self) -> str:
        """Return any extra output required at the end when logging.

        NB Override this for specific subclasses if we want extra output.

        Returns:
            str: The extra output.
        """
        return f"{self.raw!r}"

    def edit(
        self, raw: Optional[str] = None, source_fixes: Optional[List[SourceFix]] = None
    ) -> "RawSegment":
        """Create a new segment, with exactly the same position but different content.

        Args:
            raw (Optional[str]): The new content for the segment.
            source_fixes (Optional[List[SourceFix]]): A list of fixes to be applied to
                the segment.

        Returns:
            RawSegment: A copy of this object with new contents.

        Used mostly by fixes.

        NOTE: This *doesn't* copy the uuid. The edited segment is a new
        segment.

        """
        return self.__class__(
            raw=raw or self.raw,
            pos_marker=self.pos_marker,
            instance_types=self.instance_types,
            trim_start=self.trim_start,
            trim_chars=self.trim_chars,
            quoted_value=self.quoted_value,
            escape_replacements=self.escape_replacements,
            casefold=self.casefold,
            source_fixes=source_fixes or self.source_fixes,
        )

    def _get_raw_segment_kwargs(self) -> Dict[str, Any]:
        return {
            "quoted_value": self.quoted_value,
            "escape_replacements": self.escape_replacements,
            "casefold": self.casefold,
        }

    # ################ CLASS METHODS

    @classmethod
    def from_result_segments(
        cls,
        result_segments: Tuple[BaseSegment, ...],
        segment_kwargs: Dict[str, Any],
    ) -> "RawSegment":
        """Create a RawSegment from result segments."""
        assert len(result_segments) == 1
        raw_seg = cast("RawSegment", result_segments[0])
        new_segment_kwargs = raw_seg._get_raw_segment_kwargs()
        new_segment_kwargs.update(segment_kwargs)
        return cls(
            raw=raw_seg.raw,
            pos_marker=raw_seg.pos_marker,
            **new_segment_kwargs,
        )


__all__ = [
    "PositionMarker",
    "RawSegment",
    "SourceFix",
]



"""This is an example of providing config overrides."""

from sqlfluff.core import FluffConfig, Linter

sql = "SELECT 1\n"


config = FluffConfig(
    overrides={
        "dialect": "snowflake",
        # NOTE: We explicitly set the string "none" here rather
        # than a None literal so that it overrides any config
        # set by any config files in the path.
        "library_path": "none",
    }
)

linted_file = Linter(config=config).lint_string(sql)

assert linted_file.get_violations() == []



"""Test the interactions of AL09, CP02 & RF06.

AL09: Self aliasing
CP02: Identifier Capitalisation
RF06: Identifier Quoting
"""

import pytest

from sqlfluff.core import Linter

input_query = """
select
    a as A,
    B as b,
    "C" as C,
    "d" as d,
    "E" as e,
    "f" as F,
    g as "G",
    h as h,
    I as I
from foo
"""


@pytest.mark.parametrize(
    "rules,dialect,fixed_sql,post_fix_errors",
    [
        # NOTE: The first few examples here are with ANSI which is
        # configured as a natively UPPERCASE dialect.
        (
            ["AL09"],
            "ansi",
            """
select
    a as A,
    B as b,
    "C" as C,
    "d" as d,
    "E" as e,
    "f" as F,
    g as "G",
    h,
    I
from foo
""",
            [
                # These two (A & B) are detected as self aliases, but not
                # auto-fixed, because the intent is ambiguous.
                # Should the alias/reference be quoted or removed?
                ("AL09", 3, 5),
                ("AL09", 4, 5),
            ],
        ),
        (
            ["CP02"],
            "ansi",
            """
select
    a as a,
    b as b,
    "C" as c,
    "d" as d,
    "E" as e,
    "f" as f,
    g as "G",
    h as h,
    i as i
from foo
""",
            [],
        ),
        (
            ["RF06"],
            "ansi",
            """
select
    a as A,
    B as b,
    C as C,
    "d" as d,
    E as e,
    "f" as F,
    g as G,
    h as h,
    I as I
from foo
""",
            [],
        ),
        (
            ["AL09", "CP02"],
            "ansi",
            """
select
    a,
    b,
    "C" as c,
    "d" as d,
    "E" as e,
    "f" as f,
    g as "G",
    h,
    i
from foo
""",
            # NOTE: When CP02 is active, AL09 errors are no longer
            # present, because CP02 allowed them to be resolved.
            [],
        ),
        (
            ["AL09", "RF06"],
            "ansi",
            """
select
    a as A,
    B as b,
    C,
    "d" as d,
    E as e,
    "f" as F,
    g as G,
    h,
    I
from foo
""",
            [
                # Without CPO2, the errors on line 3 & 5 are present. They're
                # detected as self-aliases, but with ambiguous fixes (A & B).
                ("AL09", 3, 5),
                ("AL09", 4, 5),
                # Additionally, with RF06 removing quotes, it creates two
                # new issues, where the previously quoted aliases are now
                # unquoted, but still different cases (E & G).
                ("AL09", 7, 5),
                ("AL09", 9, 5),
            ],
        ),
        (
            ["CP02", "RF06"],
            "ansi",
            """
select
    a as a,
    b as b,
    c as c,
    "d" as d,
    e as e,
    "f" as f,
    g as g,
    h as h,
    i as i
from foo
""",
            [],
        ),
        (
            ["AL09", "CP02", "RF06"],
            "ansi",
            """
select
    a,
    b,
    c,
    "d" as d,
    e,
    "f" as f,
    g,
    h,
    i
from foo
""",
            [],
        ),
        # Postgres is natively lowercase, and so the results are
        # different.
        (
            ["AL09", "CP02", "RF06"],
            "postgres",
            """
select
    a,
    b,
    "C" as c,
    d,
    "E" as e,
    f,
    g as "G",
    h,
    i
from foo
""",
            [],
        ),
        # DuckDB is always case insensitive so likewise has a different result.
        (
            ["AL09", "CP02", "RF06"],
            "duckdb",
            """
select
    a,
    b,
    c,
    d,
    e,
    f,
    g,
    h,
    i
from foo
""",
            [],
        ),
        # Clickhouse is always case sensitive has a more conservative result.
        # All the quotes are gone, but all the aliases with a case change remain.
        (
            # NOTE: Testing without CP02 as that rule is much less appropriate
            # for clickhouse.
            ["AL09", "RF06"],
            "clickhouse",
            """
select
    a as A,
    B as b,
    C,
    d,
    E as e,
    f as F,
    g as G,
    h,
    I
from foo
""",
            # None of those aliases should be flagged as an issue.
            [],
        ),
    ],
)
def test__rules__std_AL09_CP02_RF06(rules, dialect, fixed_sql, post_fix_errors):
    """Test interactions between AL09, CP02 & RF06."""
    print(f"Running with rules: {rules}")
    linter = Linter(dialect=dialect, rules=rules)
    result = linter.lint_string(input_query, fix=True)
    fixed, _ = result.fix_string()
    assert fixed == fixed_sql
    # Check violations after fix.
    # NOTE: We should really use the rules testing utilities here
    # but they don't yet support multiple rules.
    post_fix_result = linter.lint_string(fixed, fix=False)
    assert post_fix_result.check_tuples() == post_fix_errors



"""Tests for the Sequence grammar.

NOTE: All of these tests depend somewhat on the KeywordSegment working as planned.
"""

import logging

import pytest

from sqlfluff.core.errors import SQLParseError
from sqlfluff.core.parser import Dedent, Indent, KeywordSegment, StringParser
from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar import Bracketed, Conditional, Sequence
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.types import ParseMode


def test__parser__grammar_sequence_repr():
    """Test the Sequence grammar __repr__ method."""
    bar = StringParser("bar", KeywordSegment)
    assert repr(bar) == "<StringParser: 'BAR'>"
    foo = StringParser("foo", KeywordSegment)
    sequence = Sequence(bar, foo)
    assert (
        repr(sequence) == "<Sequence: [<StringParser: 'BAR'>, <StringParser: 'FOO'>]>"
    )


def test__parser__grammar_sequence_nested_match(test_segments, caplog):
    """Test the Sequence grammar when nested."""
    bar = StringParser("bar", KeywordSegment)
    foo = StringParser("foo", KeywordSegment)
    baar = StringParser("baar", KeywordSegment)
    g = Sequence(Sequence(bar, foo), baar)

    ctx = ParseContext(dialect=None)
    # Confirm the structure of the test segments:
    assert [s.raw for s in test_segments] == ["bar", " \t ", "foo", "baar", " \t ", ""]

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
        # Matching just the start of the list shouldn't work.
        result1 = g.match(test_segments[:3], 0, ctx)

    assert not result1  # Check it returns falsy

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
        # Matching the whole list should.
        result2 = g.match(test_segments, 0, ctx)

    assert result2  # Check it returns truthy
    assert result2 == MatchResult(
        matched_slice=slice(0, 4),  # NOTE: One of these is space.
        child_matches=(
            MatchResult(
                matched_slice=slice(0, 1),
                matched_class=KeywordSegment,
                segment_kwargs={"instance_types": ("keyword",)},
            ),
            MatchResult(
                matched_slice=slice(2, 3),
                matched_class=KeywordSegment,
                segment_kwargs={"instance_types": ("keyword",)},
            ),
            MatchResult(
                matched_slice=slice(3, 4),
                matched_class=KeywordSegment,
                segment_kwargs={"instance_types": ("keyword",)},
            ),
        ),
    )


@pytest.mark.parametrize(
    "mode,sequence,terminators,input_slice,output_tuple",
    [
        # #####
        # Test matches where we should get something, and that's
        # the whole sequence.
        # NOTE: Include a little whitespace in the slice (i.e. the first _two_
        # segments) to check that it isn't included in the match.
        (ParseMode.STRICT, ["a"], [], slice(None, 2), (("keyword", "a"),)),
        (ParseMode.GREEDY, ["a"], [], slice(None, 2), (("keyword", "a"),)),
        (ParseMode.GREEDY_ONCE_STARTED, ["a"], [], slice(None, 2), (("keyword", "a"),)),
        # #####
        # Test matching on sequences where we run out of segments before matching
        # the whole sequence.
        # STRICT returns no match.
        (ParseMode.STRICT, ["a", "b"], [], slice(None, 2), ()),
        # GREEDY & GREEDY_ONCE_STARTED returns the content as unparsable, and
        # still don't include the trailing whitespace. The return value does
        # however have the matched "a" as a keyword and not a raw.
        (
            ParseMode.GREEDY,
            ["a", "b"],
            [],
            slice(None, 2),
            (("unparsable", (("keyword", "a"),)),),
        ),
        (
            ParseMode.GREEDY_ONCE_STARTED,
            ["a", "b"],
            [],
            slice(None, 2),
            (("unparsable", (("keyword", "a"),)),),
        ),
        # #####
        # Test matching on sequences where we fail to match the first element.
        # STRICT & GREEDY_ONCE_STARTED return no match.
        (ParseMode.STRICT, ["b"], [], slice(None, 2), ()),
        (ParseMode.GREEDY_ONCE_STARTED, ["b"], [], slice(None, 2), ()),
        # GREEDY claims the remaining elements (unmutated) as unparsable, but
        # does not claim any trailing whitespace.
        (
            ParseMode.GREEDY,
            ["b"],
            [],
            slice(None, 2),
            (("unparsable", (("raw", "a"),)),),
        ),
        # #####
        # Test matches where we should match the sequence fully, but there's more
        # to match.
        # First without terminators...
        # STRICT ignores the rest.
        (ParseMode.STRICT, ["a"], [], slice(None, 5), (("keyword", "a"),)),
        # The GREEDY modes claim the rest as unparsable.
        # NOTE: the whitespace in between is _not_ unparsable.
        (
            ParseMode.GREEDY,
            ["a"],
            [],
            slice(None, 5),
            (
                ("keyword", "a"),
                ("whitespace", " "),
                ("unparsable", (("raw", "b"), ("whitespace", " "), ("raw", "c"))),
            ),
        ),
        (
            ParseMode.GREEDY_ONCE_STARTED,
            ["a"],
            [],
            slice(None, 5),
            (
                ("keyword", "a"),
                ("whitespace", " "),
                ("unparsable", (("raw", "b"), ("whitespace", " "), ("raw", "c"))),
            ),
        ),
        # Second *with* terminators.
        # NOTE: The whitespace before the terminator is not included.
        (ParseMode.STRICT, ["a"], ["c"], slice(None, 5), (("keyword", "a"),)),
        (
            ParseMode.GREEDY,
            ["a"],
            ["c"],
            slice(None, 5),
            (("keyword", "a"), ("whitespace", " "), ("unparsable", (("raw", "b"),))),
        ),
        (
            ParseMode.GREEDY_ONCE_STARTED,
            ["a"],
            ["c"],
            slice(None, 5),
            (("keyword", "a"), ("whitespace", " "), ("unparsable", (("raw", "b"),))),
        ),
        # #####
        # Test matches where we match the first element of a sequence but not the
        # second (with terminators)
        (ParseMode.STRICT, ["a", "x"], ["c"], slice(None, 5), ()),
        # NOTE: For GREEDY modes, the matched portion is not included as an "unparsable"
        # only the portion which failed to match. The terminator is not included and
        # the matched portion is still mutated correctly.
        (
            ParseMode.GREEDY,
            ["a", "x"],
            ["c"],
            slice(None, 5),
            (("keyword", "a"), ("whitespace", " "), ("unparsable", (("raw", "b"),))),
        ),
        (
            ParseMode.GREEDY_ONCE_STARTED,
            ["a", "x"],
            ["c"],
            slice(None, 5),
            (("keyword", "a"), ("whitespace", " "), ("unparsable", (("raw", "b"),))),
        ),
        # #####
        # Test competition between sequence elements and terminators.
        # In GREEDY_ONCE_STARTED, the first element is matched before any terminators.
        (
            ParseMode.GREEDY_ONCE_STARTED,
            ["a"],
            ["a"],
            slice(None, 2),
            (("keyword", "a"),),
        ),
        # In GREEDY, the terminator is matched first and so takes precedence.
        (
            ParseMode.GREEDY,
            ["a"],
            ["a"],
            slice(None, 2),
            (),
        ),
        # NOTE: In these last two cases, the "b" isn't included because it acted as
        # a terminator before being considered in the sequence.
        (
            ParseMode.GREEDY_ONCE_STARTED,
            ["a", "b"],
            ["b"],
            slice(None, 3),
            (("unparsable", (("keyword", "a"),)),),
        ),
        (
            ParseMode.GREEDY,
            ["a", "b"],
            ["b"],
            slice(None, 3),
            (("unparsable", (("keyword", "a"),)),),
        ),
    ],
)
def test__parser__grammar_sequence_modes(
    mode,
    sequence,
    terminators,
    input_slice,
    output_tuple,
    structural_parse_mode_test,
):
    """Test the Sequence grammar with various parse modes.

    In particular here we're testing the treatment of unparsable
    sections.
    """
    structural_parse_mode_test(
        ["a", " ", "b", " ", "c", "d", " ", "d"],
        Sequence,
        sequence,
        terminators,
        {},
        mode,
        input_slice,
        output_tuple,
    )


@pytest.mark.parametrize(
    "input_seed,mode,sequence,kwargs,output_tuple",
    [
        # A sequence that isn't bracketed shouldn't match.
        # Regardless of mode.
        (["a"], ParseMode.STRICT, ["a"], {}, ()),
        (["a"], ParseMode.GREEDY, ["a"], {}, ()),
        # Test potential empty brackets (no whitespace)
        (
            ["(", ")"],
            ParseMode.STRICT,
            [],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        (
            ["(", ")"],
            ParseMode.GREEDY,
            [],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        # Test potential empty brackets (with whitespace)
        (
            ["(", " ", ")"],
            ParseMode.STRICT,
            [],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("whitespace", " "),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        (
            ["(", " ", ")"],
            ParseMode.GREEDY,
            [],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("whitespace", " "),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        (
            ["(", " ", ")"],
            ParseMode.STRICT,
            [],
            # Strict matching, without allowing gaps, shouldn't match.
            {"allow_gaps": False},
            (),
        ),
        (
            ["(", " ", ")"],
            ParseMode.GREEDY,
            [],
            # Greedy matching, without allowing gaps, should return unparsable.
            # NOTE: This functionality doesn't get used much.
            {"allow_gaps": False},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("unparsable", (("whitespace", " "),)),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        # Happy path content match.
        (
            ["(", "a", ")"],
            ParseMode.STRICT,
            ["a"],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("keyword", "a"),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        # Content match fails
        (
            ["(", "a", ")"],
            ParseMode.STRICT,
            ["b"],
            {},
            (),
        ),
        (
            ["(", "a", ")"],
            ParseMode.GREEDY,
            ["b"],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("unparsable", (("raw", "a"),)),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        # Partial matches (not whole grammar matched)
        (
            ["(", "a", ")"],
            ParseMode.STRICT,
            ["a", "b"],
            {},
            (),
        ),
        (
            ["(", "a", ")"],
            ParseMode.GREEDY,
            ["a", "b"],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("unparsable", (("keyword", "a"),)),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        # Partial matches (not whole sequence matched)
        (
            ["(", "a", " ", "b", ")"],
            ParseMode.STRICT,
            ["a"],
            {},
            (),
        ),
        (
            ["(", "a", " ", "b", ")"],
            ParseMode.GREEDY,
            ["a"],
            {},
            (
                (
                    "bracketed",
                    (
                        ("start_bracket", "("),
                        ("indent", ""),
                        ("keyword", "a"),
                        ("whitespace", " "),
                        ("unparsable", (("raw", "b"),)),
                        ("dedent", ""),
                        ("end_bracket", ")"),
                    ),
                ),
            ),
        ),
        # Test an unwrapped path (with square brackets)
        (
            ["[", "a", " ", "b", "]"],
            ParseMode.GREEDY,
            ["a"],
            {"bracket_type": "square"},
            (
                ("start_square_bracket", "["),
                ("indent", ""),
                ("keyword", "a"),
                ("whitespace", " "),
                ("unparsable", (("raw", "b"),)),
                ("dedent", ""),
                ("end_square_bracket", "]"),
            ),
        ),
    ],
)
def test__parser__grammar_bracketed_modes(
    input_seed,
    mode,
    sequence,
    kwargs,
    output_tuple,
    structural_parse_mode_test,
):
    """Test the Bracketed grammar with various parse modes."""
    structural_parse_mode_test(
        input_seed,
        Bracketed,
        sequence,
        [],
        kwargs,
        mode,
        slice(None, None),
        output_tuple,
    )


@pytest.mark.parametrize(
    "input_seed,mode,sequence",
    [
        # Unclosed brackets always raise errors.
        (["(", "a"], ParseMode.STRICT, ["a"]),
        (["(", "a"], ParseMode.GREEDY, ["a"]),
    ],
)
def test__parser__grammar_bracketed_error_modes(
    input_seed,
    mode,
    sequence,
    structural_parse_mode_test,
):
    """Test the Bracketed grammar with various parse modes."""
    with pytest.raises(SQLParseError):
        structural_parse_mode_test(
            input_seed,
            Bracketed,
            sequence,
            [],
            {},
            mode,
            slice(None, None),
            (),
        )


def test__parser__grammar_sequence_indent_conditional_match(test_segments, caplog):
    """Test the Sequence grammar with indents."""
    bar = StringParser("bar", KeywordSegment)
    foo = StringParser("foo", KeywordSegment)
    # We will assume the default config has indented_joins = False.
    # We're testing without explicitly setting the `config_type` because
    # that's the assumed way of using the grammar in practice.
    g = Sequence(
        Dedent,
        Conditional(Indent, indented_joins=False),
        bar,
        Conditional(Indent, indented_joins=True),
        foo,
        Dedent,
    )
    ctx = ParseContext(dialect=None)
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.parser"):
        m = g.match(test_segments, 0, parse_context=ctx)

    assert m == MatchResult(
        matched_slice=slice(0, 3),  # NOTE: One of these is space.
        child_matches=(
            # The two child keywords
            MatchResult(
                matched_slice=slice(0, 1),
                matched_class=KeywordSegment,
                segment_kwargs={"instance_types": ("keyword",)},
            ),
            MatchResult(
                matched_slice=slice(2, 3),
                matched_class=KeywordSegment,
                segment_kwargs={"instance_types": ("keyword",)},
            ),
        ),
        insert_segments=(
            (0, Dedent),  # The starting, unconditional dedent.
            (0, Indent),  # The conditional (activated) Indent.
            # NOTE: There *isn't* the other Indent.
            (3, Dedent),  # The closing unconditional dedent.
            # NOTE: This last one is still included even though it's
            # after the last matched segment.
        ),
    )



"""Implementation of Rule JJ01."""

from typing import List, Tuple

from sqlfluff.core.parser.segments import BaseSegment, SourceFix
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import RootOnlyCrawler
from sqlfluff.core.templaters import JinjaTemplater


class Rule_JJ01(BaseRule):
    """Jinja tags should have a single whitespace on either side.

    This rule is only active if the ``jinja`` templater (or one of it's
    subclasses, like the ``dbt`` templater) are used for the current file.

    **Anti-pattern**

    Jinja tags with either no whitespace or very long whitespace
    are hard to read.

    .. code-block:: jinja
       :force:

        SELECT {{    a     }} from {{ref('foo')}}

    **Best practice**

    A single whitespace surrounding Jinja tags, alternatively
    longer gaps containing newlines are acceptable.

    .. code-block:: jinja
       :force:

        SELECT {{ a }} from {{ ref('foo') }};
        SELECT {{ a }} from {{
            ref('foo')
        }};
    """

    name = "jinja.padding"
    aliases = ("L046",)
    groups = ("all", "core", "jinja")
    crawl_behaviour = RootOnlyCrawler()
    targets_templated = True
    is_fix_compatible = True

    @staticmethod
    def _get_whitespace_ends(s: str) -> Tuple[str, str, str, str, str]:
        """Remove tag ends and partition off any whitespace ends.

        This function assumes that we've already trimmed the string
        to just the tag, and will raise an AssertionError if not.
        >>> Rule_JJ01._get_whitespace_ends('  {{not_trimmed}}   ')
        Traceback (most recent call last):
            ...
        AssertionError

        In essence it divides up a tag into the end tokens, any
        leading or trailing whitespace and the inner content
        >>> Rule_JJ01._get_whitespace_ends('{{ my_content }}')
        ('{{', ' ', 'my_content', ' ', '}}')

        It also works with block tags and more complicated content
        and end markers.
        >>> Rule_JJ01._get_whitespace_ends('{%+if a + b is True     -%}')
        ('{%+', '', 'if a + b is True', '     ', '-%}')
        """
        assert s[0] == "{" and s[-1] == "}"
        # Jinja tags all have a length of two. We can use slicing
        # to remove them easily.
        main = s[2:-2]
        pre = s[:2]
        post = s[-2:]
        # Optionally Jinja tags may also have plus of minus notation
        # https://jinja2docs.readthedocs.io/en/stable/templates.html#whitespace-control
        modifier_chars = ["+", "-"]
        if main and main[0] in modifier_chars:
            main = main[1:]
            pre = s[:3]
        if main and main[-1] in modifier_chars:
            main = main[:-1]
            post = s[-3:]
        inner = main.strip()
        pos = main.find(inner)
        return pre, main[:pos], inner, main[pos + len(inner) :], post

    @classmethod
    def _find_raw_at_src_idx(cls, segment: BaseSegment, src_idx: int):
        """Recursively search to find a raw segment for a position in the source.

        NOTE: This assumes it's not being called on a `raw`.

        In the case that there are multiple potential targets, we will find the
        first.
        """
        assert segment.segments
        for seg in segment.segments:
            if not seg.pos_marker:  # pragma: no cover
                continue
            src_slice = seg.pos_marker.source_slice
            # If it's before, skip onward.
            if src_slice.stop <= src_idx:
                continue
            # Is the current segment raw?
            if seg.is_raw():
                return seg
            # Otherwise recurse
            return cls._find_raw_at_src_idx(seg, src_idx)

    def _eval(self, context: RuleContext) -> List[LintResult]:
        """Look for non-literal segments.

        NOTE: The existing crawlers don't filter very well for only templated
        code, and so we process the whole file from the root here.
        """
        # If the position maker for the root segment is literal then there's
        # no templated code. So we can return early.
        assert context.segment.pos_marker
        if context.segment.pos_marker.is_literal():
            return []

        # We'll need the templated file. If for whatever reason it's
        # not present, abort.
        if not context.templated_file:  # pragma: no cover
            return []

        # We also only work with setups which use the jinja templater
        # or a derivative of that. Otherwise return empty.
        # NOTE: The `templater_obj` is not available in parallel operations
        # and we don't really want to rehydrate a templater just to check
        # what type it is, so use `get_templater_class()`.
        _templater_class = context.config.get_templater_class()
        if not issubclass(_templater_class, JinjaTemplater):
            self.logger.debug(f"Detected non-jinja templater: {_templater_class.name}")
            return []

        results = []
        # Work through the templated slices
        for raw_slice in context.templated_file.raw_sliced:
            # We only want templated slices.
            if raw_slice.slice_type not in ("templated", "block_start", "block_end"):
                continue

            stripped = raw_slice.raw.strip()
            if not stripped or stripped[0] != "{" or stripped[-1] != "}":
                continue  # pragma: no cover

            self.logger.debug(
                "Tag found @ source index %s: %r ", raw_slice.source_idx, stripped
            )

            # Partition and Position
            src_idx = raw_slice.source_idx
            tag_pre, ws_pre, inner, ws_post, tag_post = self._get_whitespace_ends(
                stripped
            )
            position = raw_slice.raw.find(stripped[0])

            self.logger.debug(
                "Tag string segments: %r | %r | %r | %r | %r @ %s + %s",
                tag_pre,
                ws_pre,
                inner,
                ws_post,
                tag_post,
                src_idx,
                position,
            )

            # For the following section, whitespace should be a single
            # whitespace OR it should contain a newline.

            pre_fix = None
            post_fix = None
            # Check the initial whitespace.
            if not ws_pre or (ws_pre != " " and "\n" not in ws_pre):
                pre_fix = " "
            # Check latter whitespace.
            if not ws_post or (ws_post != " " and "\n" not in ws_post):
                post_fix = " "

            # If no fixes, continue
            if pre_fix is None and post_fix is None:
                continue

            fixed = (
                tag_pre + (pre_fix or ws_pre) + inner + (post_fix or ws_post) + tag_post
            )

            # We need to identify a raw segment to attach to fix to.
            raw_seg = self._find_raw_at_src_idx(context.segment, src_idx)

            # If that raw segment already has fixes, don't apply it again.
            # We're likely on a second pass.
            if raw_seg.source_fixes:
                continue

            source_fixes = [
                SourceFix(
                    fixed,
                    slice(
                        src_idx + position,
                        src_idx + position + len(stripped),
                    ),
                    # This position in the templated file is rough, but
                    # close enough for sequencing.
                    raw_seg.pos_marker.templated_slice,
                )
            ]

            results.append(
                LintResult(
                    anchor=raw_seg,
                    description=f"Jinja tags should have a single "
                    f"whitespace on either side: {stripped}",
                    fixes=[
                        LintFix.replace(
                            raw_seg,
                            [raw_seg.edit(source_fixes=source_fixes)],
                        )
                    ],
                )
            )

        return results



"""Module used to test filters within the jinja template."""

from __future__ import annotations

import datetime


# https://github.com/apache/airflow/blob/main/airflow/templates.py#L50
def ds_filter(value: datetime.date | datetime.time | None) -> str | None:
    """Date filter."""
    if value is None:
        return None
    return value.strftime("%Y-%m-%d")


SQLFLUFF_JINJA_FILTERS = {"ds": ds_filter}

now = datetime.datetime(
    2006, 1, 2, 3, 4, 5, 0, tzinfo=datetime.timezone(-datetime.timedelta(hours=7))
)



"""Generate some documentation automatically.

This script generates partial documentation sections (i.e. the content of
`/docs/source/_partials/`) by importing SQLFluff and extracting data about
rules and dialects.

It should run before every docs generation so that those partial .rst files
can then be correctly referenced by other sections of the docs. For example
this file builds the file `/docs/source/_partials/rule_summaries.rst`, which
is then inserted into `/docs/source/reference/rules.rst` using the directive
`.. include:: ../_partials/rule_summaries.rst`.

This script is referenced in the `Makefile` and the `make.bat` file to ensure
it is run at the appropriate moment.
"""

import json
from collections import defaultdict
from pathlib import Path

import sqlfluff
from sqlfluff.core.plugin.host import get_plugin_manager

base_path = Path(__file__).parent.absolute()

##########################################
# Generate rule documentation dynamically.
##########################################

autogen_header = """..
    NOTE: This file is generated by the conf.py script.
    Don't edit this by hand


"""

table_header = f"""
+{'-' * 42}+{'-' * 50}+{'-' * 30}+{'-' * 20}+
|{'Bundle' : <42}|{'Rule Name' : <50}|{'Code' : <30}|{'Aliases' : <20}|
+{'=' * 42}+{'=' * 50}+{'=' * 30}+{'=' * 20}+
"""

# Extract all the rules.
print("Rule Docs Generation: Reading Rules...")
rule_bundles = defaultdict(list)
rule_list = []
for plugin_rules in get_plugin_manager().hook.get_rules():
    for rule in plugin_rules:
        _bundle_name = rule.name.split(".")[0]
        rule_bundles[_bundle_name].append(rule)
        rule_list.append((rule.code, rule.name))

# Write them into a json file for use by redirects.
print("Rule Docs Generation: Writing Rule JSON...")
with open(base_path / "source/_partials/rule_list.json", "w", encoding="utf8") as f:
    json.dump(rule_list, f)

# Write them into the table. Bundle by bundle.
print("Rule Docs Generation: Writing Rule Table...")
with open(base_path / "source/_partials/rule_table.rst", "w", encoding="utf8") as f:
    f.write(autogen_header)
    f.write(table_header)
    for bundle in sorted(rule_bundles.keys()):
        # Set the bundle name to the ref.
        _bundle_name = f":ref:`bundle_{bundle}`"
        for idx, rule in enumerate(rule_bundles[bundle]):
            step = 1  # The number of aliases per line.
            aliases = ", ".join(rule.aliases[:step]) + (
                "," if len(rule.aliases) > step else ""
            )
            name_ref = f":sqlfluff:ref:`{rule.name}`"
            code_ref = f":sqlfluff:ref:`{rule.code}`"
            f.write(
                f"| {_bundle_name : <40} | {name_ref : <48} "
                f"| {code_ref : <28} | {aliases : <18} |\n"
            )

            j = 1

            while True:
                if not rule.aliases[j:]:
                    break
                aliases = ", ".join(rule.aliases[j : j + step]) + (
                    "," if len(rule.aliases[j:]) > step else ""
                )
                f.write(f"|{' ' * 42}|{' ' * 50}|{' ' * 30}| {aliases : <18} |\n")
                j += step

            if idx + 1 < len(rule_bundles[bundle]):
                f.write(f"|{' ' * 42}+{'-' * 50}+{'-' * 30}+{'-' * 20}+\n")
            else:
                f.write(f"+{'-' * 42}+{'-' * 50}+{'-' * 30}+{'-' * 20}+\n")
            # Unset the bundle name so we don't repeat it.
            _bundle_name = ""
    f.write("\n\n")


# Write each of the summary files.
print("Rule Docs Generation: Writing Rule Summaries...")
with open(base_path / "source/_partials/rule_summaries.rst", "w", encoding="utf8") as f:
    f.write(autogen_header)
    for bundle in sorted(rule_bundles.keys()):
        if "sql" in bundle:
            # This accounts for things like "TSQL"
            header_name = bundle.upper()
        else:
            header_name = bundle.capitalize()
        # Write the bundle header.
        f.write(
            f".. _bundle_{bundle}:\n\n"
            f"{header_name} bundle\n"
            f"{'-' * (len(bundle) + 7)}\n\n"
        )
        for rule in rule_bundles[bundle]:
            f.write(
                f".. sqlfluff:rule:: {rule.code}\n"
                f"                   {rule.name}\n\n"
            )
            # Separate off the heading so we can bold it.
            heading, _, doc_body = rule.__doc__.partition("\n")
            underline_char = '"'
            f.write(f"    {heading}\n")
            f.write(f"    {underline_char * len(heading)}\n\n")
            f.write("    " + doc_body)
            f.write("\n\n")

print("Rule Docs Generation: Done")

# Extract all the dialects.
print("Dialect Docs Generation: Reading Dialects...")
# We make a dictionary of all of them first, because we want to force the ANSI
# one to be first.
dialect_dict = {dialect.label: dialect for dialect in sqlfluff.list_dialects()}
dialect_list = [dialect_dict["ansi"]] + [
    dialect for dialect_name, dialect in dialect_dict.items() if dialect_name != "ansi"
]

# Write each of the summary files.
print("Dialect Docs Generation: Writing Dialect Summaries...")
with open(
    base_path / "source/_partials/dialect_summaries.rst", "w", encoding="utf8"
) as f:
    f.write(autogen_header)
    for dialect in dialect_list:
        f.write(
            f".. _{dialect.label}_dialect_ref:\n\n"
            f"{dialect.name}\n{'-' * len(dialect.name)}\n\n"
            f"**Label**: ``{dialect.label}``\n\n"
        )
        if dialect.label != "ansi":
            f.write(
                f"**Inherits from**: :ref:`{dialect.inherits_from}_dialect_ref`\n\n"
            )
        if dialect.docstring:
            f.write(dialect.docstring + "\n\n")



"""Tests for reindenting methods.

Specifically:
- ReflowPoint.indent_to()
- ReflowPoint.get_indent()
- deduce_line_indent()
"""

import logging
import sys
from typing import Callable, List, Tuple, Type

import pytest

from sqlfluff.core import FluffConfig, Linter
from sqlfluff.core.linter.fix import apply_fixes, compute_anchor_edit_info
from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.plugin.host import get_plugin_manager, purge_plugin_manager
from sqlfluff.core.templaters import RawTemplater
from sqlfluff.core.templaters.base import RawFileSlice, TemplatedFileSlice
from sqlfluff.core.templaters.jinja import JinjaTemplater
from sqlfluff.utils.reflow.helpers import deduce_line_indent, fixes_from_results
from sqlfluff.utils.reflow.reindent import (
    _crawl_indent_points,
    _IndentLine,
    _IndentPoint,
    lint_indent_points,
)
from sqlfluff.utils.reflow.sequence import ReflowSequence


def parse_ansi_string(sql, config):
    """Parse an ansi sql string for testing."""
    linter = Linter(config=config)
    return linter.parse_string(sql).tree


class SpecialMarkerInserter(JinjaTemplater):
    """Inserts special marker slices in a sliced file.

    Some templater plugins might insert custom marker slices that are of zero source
    string length, including an empty source string.  This mock templater simulates
    this behavior by adding a marker slice like this after every block_start slice.
    """

    name = "special_marker_inserter"

    def slice_file(
        self, raw_str: str, render_func: Callable[[str], str], config=None
    ) -> Tuple[List[RawFileSlice], List[TemplatedFileSlice], str]:
        """Patch a sliced file returned by the superclass."""
        raw_sliced, sliced_file, templated_str = super().slice_file(
            raw_str, render_func, config
        )

        patched_sliced_file = []
        for templated_slice in sliced_file:
            patched_sliced_file.append(templated_slice)
            # Add an EMPTY special_marker slice after every block_start.
            if templated_slice.slice_type == "block_start":
                # Note that both the source_slice AND the templated_slice are empty.
                source_pos = templated_slice.source_slice.stop
                templated_pos = templated_slice.templated_slice.stop
                patched_sliced_file.append(
                    TemplatedFileSlice(
                        "special_marker",
                        slice(source_pos, source_pos),
                        slice(templated_pos, templated_pos),
                    )
                )

        return raw_sliced, patched_sliced_file, templated_str


@hookimpl
def get_templaters() -> List[Type[RawTemplater]]:
    """Return templaters provided by this test module."""
    return [SpecialMarkerInserter]


@pytest.mark.parametrize(
    "raw_sql_in,elem_idx,indent_to,point_sql_out",
    [
        # Trivial Case
        ("select\n  1", 1, "  ", "\n  "),
        # Change existing indents
        ("select\n  1", 1, "    ", "\n    "),
        ("select\n  1", 1, " ", "\n "),
        ("select\n1", 1, "  ", "\n  "),
        ("select\n  1", 1, "", "\n"),
        # Create new indents
        ("select 1", 1, "  ", "\n  "),
        ("select 1", 1, " ", "\n "),
        ("select 1", 1, "", "\n"),
        ("select      1", 1, "  ", "\n  "),
    ],
)
def test_reflow__point_indent_to(
    raw_sql_in, elem_idx, indent_to, point_sql_out, default_config, caplog
):
    """Test the ReflowPoint.indent_to() method directly."""
    root = parse_ansi_string(raw_sql_in, default_config)
    print(root.stringify())
    seq = ReflowSequence.from_root(root, config=default_config)
    elem = seq.elements[elem_idx]
    print("Element: ", elem)

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules.reflow"):
        new_fixes, new_point = elem.indent_to(
            indent_to,
            before=seq.elements[elem_idx - 1].segments[-1],
            after=seq.elements[elem_idx + 1].segments[0],
        )

    print(new_fixes)
    assert new_point.raw == point_sql_out


@pytest.mark.parametrize(
    "raw_sql_in,elem_idx,indent_out",
    [
        # Null case
        ("select 1", 1, None),
        # Trivial Case
        ("select\n  1", 1, "  "),
        # Harder Case (i.e. take the last indent)
        ("select\n \n  \n   1", 1, "   "),
    ],
)
def test_reflow__point_get_indent(
    raw_sql_in, elem_idx, indent_out, default_config, caplog
):
    """Test the ReflowPoint.get_indent() method directly."""
    root = parse_ansi_string(raw_sql_in, default_config)
    print(root.stringify())
    seq = ReflowSequence.from_root(root, config=default_config)
    elem = seq.elements[elem_idx]
    print("Element: ", elem)

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules.reflow"):
        result = elem.get_indent()

    assert result == indent_out


@pytest.mark.parametrize(
    "raw_sql_in,target_raw,indent_out",
    [
        # Trivial case
        ("select 1", "select", ""),
        ("select 1", "1", ""),
        # Easy Case
        ("select\n  1", "1", "  "),
        # Harder Cases (i.e. take the last indent)
        ("select\n \n  \n   1", "1", "   "),
        ("select\n \n  \n   1+2+3+4", "4", "   "),
        ("select\n   1 + 2", "2", "   "),
    ],
)
def test_reflow__deduce_line_indent(
    raw_sql_in, target_raw, indent_out, default_config, caplog
):
    """Test the deduce_line_indent() method directly."""
    root = parse_ansi_string(raw_sql_in, default_config)
    print(root.stringify())

    for target_seg in root.raw_segments:
        if target_seg.raw == target_raw:
            break
    else:
        raise ValueError("Target Raw Not Found")
    print("Target: ", target_seg)

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules.reflow"):
        result = deduce_line_indent(target_seg, root)

    assert result == indent_out


@pytest.mark.parametrize(
    "raw_sql_in,templater,points_out",
    [
        # Trivial
        (
            "select 1",
            "raw",
            [
                # No point at the start.
                # Point after select (not newline)
                _IndentPoint(
                    idx=1,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=None,
                    is_line_break=False,
                    untaken_indents=(),
                ),
                # Point after 1 (not newline either)
                _IndentPoint(
                    idx=3,
                    indent_impulse=-1,
                    indent_trough=-1,
                    initial_indent_balance=1,
                    last_line_break_idx=None,
                    is_line_break=False,
                    untaken_indents=(1,),
                ),
            ],
        ),
        (
            "\nselect 1\n",
            "raw",
            [
                # Start point
                _IndentPoint(
                    idx=0,
                    indent_impulse=0,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=None,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                # Point after select (not newline)
                _IndentPoint(
                    idx=2,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=0,
                    is_line_break=False,
                    untaken_indents=(),
                ),
                # Point after 1 (is newline)
                _IndentPoint(
                    idx=4,
                    indent_impulse=-1,
                    indent_trough=-1,
                    initial_indent_balance=1,
                    last_line_break_idx=0,
                    is_line_break=True,
                    untaken_indents=(1,),
                ),
            ],
        ),
        (
            "select\n1",
            "raw",
            [
                # No point at the start.
                # Point after select (not newline)
                _IndentPoint(
                    idx=1,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=None,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                # Point after 1 (is not newline)
                _IndentPoint(
                    idx=3,
                    indent_impulse=-1,
                    indent_trough=-1,
                    initial_indent_balance=1,
                    last_line_break_idx=1,
                    is_line_break=False,
                    untaken_indents=(),
                ),
            ],
        ),
        # More stretching cases.
        (
            "SELECT\n    r.a,\n    s.b\nFROM r\nJOIN s\n    "
            "ON\n        r.a = s.a\n        AND true",
            "raw",
            [
                # No point at the start.
                # After SELECT
                _IndentPoint(
                    idx=1,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=None,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                _IndentPoint(
                    idx=9,
                    indent_impulse=0,
                    indent_trough=0,
                    initial_indent_balance=1,
                    last_line_break_idx=1,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                # Before FROM
                _IndentPoint(
                    idx=15,
                    indent_impulse=-1,
                    indent_trough=-1,
                    initial_indent_balance=1,
                    last_line_break_idx=9,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                # Untaken indent before "r"
                _IndentPoint(
                    idx=17,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=15,
                    is_line_break=False,
                    untaken_indents=(),
                ),
                # Before JOIN (-1 balance to take us back to
                # baseline (in line with FROM))
                # NOTE: It keeps the untaken indent from the
                # previous point, but shouldn't use it.
                _IndentPoint(
                    idx=19,
                    indent_impulse=-1,
                    indent_trough=-1,
                    initial_indent_balance=1,
                    last_line_break_idx=15,
                    is_line_break=True,
                    untaken_indents=(1,),
                ),
                # Untaken indent before "s"
                _IndentPoint(
                    idx=21,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=0,
                    last_line_break_idx=19,
                    is_line_break=False,
                    untaken_indents=(),
                ),
                # NOTE: this is an interesting one. It's a Dedent-Indent pair.
                # There's a zero balance, and a trough of -1. We carry in the previous
                # untaken indent. But should pass if forward after this.
                _IndentPoint(
                    idx=23,
                    indent_impulse=0,
                    indent_trough=-1,
                    initial_indent_balance=1,
                    last_line_break_idx=19,
                    is_line_break=True,
                    untaken_indents=(1,),
                ),
                # After ON. Default is indented_on_contents = True, so there is
                # an indent here. We *SHOULDN'T* have an untaken indent here,
                # because while there was one at the last point, the trough
                # of the last point should have cleared it.
                _IndentPoint(
                    idx=25,
                    indent_impulse=1,
                    indent_trough=0,
                    initial_indent_balance=1,
                    last_line_break_idx=23,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                # Before AND
                _IndentPoint(
                    idx=39,
                    indent_impulse=0,
                    indent_trough=0,
                    initial_indent_balance=2,
                    last_line_break_idx=25,
                    is_line_break=True,
                    untaken_indents=(),
                ),
                # after "true"
                _IndentPoint(
                    idx=43,
                    indent_impulse=-2,
                    indent_trough=-2,
                    initial_indent_balance=2,
                    last_line_break_idx=39,
                    is_line_break=False,
                    untaken_indents=(),
                ),
            ],
        ),
        (
            "SELECT *\nFROM t1\nJOIN t2 ON true\nAND true",
            "raw",
            [
                # No point at the start.
                # NOTE: Abbreviated notation given much is the same as above.
                # After SELECT
                _IndentPoint(1, 1, 0, 0, None, False, ()),
                _IndentPoint(3, -1, -1, 1, None, True, (1,)),
                _IndentPoint(5, 1, 0, 0, 3, False, ()),
                _IndentPoint(7, -1, -1, 1, 3, True, (1,)),
                # JOIN
                _IndentPoint(9, 1, 0, 0, 7, False, ()),
                # TRICKY POINT (we're between "t2" and "ON").
                # The indent between Join and t2 wasn't taken, but we're
                # also climbing down from that here. It should be in the
                # untaken indents _here_ but not passed forward. There is
                # however another indent opportunity here which ALSO isn't
                # taken, so that one *should* be passed forward.
                _IndentPoint(11, 0, -1, 1, 7, False, (1,)),
                # TRICKY POINT (we're between "ON" and "true").
                # Default is indented_on_contents = True.
                # This means that there is an additional indent here.
                # It's not taken though. The incoming balance of 1
                # isn't taken yet either (hence a 1 in the untaken indent).
                _IndentPoint(13, 1, 0, 1, 7, False, (1,)),
                # Between "true" and "AND".
                # Balance is 2, but both untaken.
                _IndentPoint(15, 0, 0, 2, 7, True, (1, 2)),
                # End point
                _IndentPoint(19, -2, -2, 2, 15, False, (1, 2)),
            ],
        ),
        # Trailing comment case: delays indent until after the comment
        (
            "SELECT -- comment\n    1;",
            "raw",
            [
                # No point at the start.
                # After SELECT
                _IndentPoint(1, 0, 0, 0, None, False, ()),
                # After comment
                _IndentPoint(3, 1, 0, 0, None, True, ()),
                # After 1
                _IndentPoint(5, -1, -1, 1, 3, False, ()),
                # After ;
                _IndentPoint(7, 0, 0, 0, 3, False, ()),
            ],
        ),
        # Two trailing comments
        (
            "SELECT /* first comment */ /* second comment */\n    1;",
            "raw",
            [
                # No point at the start.
                # After SELECT
                _IndentPoint(1, 0, 0, 0, None, False, ()),
                # After first comment
                _IndentPoint(3, 0, 0, 0, None, False, ()),
                # After second comment
                _IndentPoint(5, 1, 0, 0, None, True, ()),
                # After 1
                _IndentPoint(7, -1, -1, 1, 5, False, ()),
                # After ;
                _IndentPoint(9, 0, 0, 0, 5, False, ()),
            ],
        ),
        # Templated case
        (
            "SELECT\n"
            "    {{ 'a' }}\n"
            "    {% for c in ['d', 'e'] %}\n"
            "    ,{{ c }}_val\n"
            "    {% endfor %}\n",
            "jinja",
            [
                # No initial indent (this is the first newline).
                _IndentPoint(1, 1, 0, 0, None, True, ()),
                # point after a
                _IndentPoint(3, 0, 0, 1, 1, True, ()),
                # point after for
                _IndentPoint(5, 1, 0, 1, 3, True, ()),
                # point after d_val
                _IndentPoint(9, -1, -1, 2, 5, True, ()),
                # point after loop
                _IndentPoint(11, 1, 0, 1, 9, True, ()),
                # point after e_val
                _IndentPoint(15, -2, -2, 2, 11, True, ()),
                # point after endfor
                _IndentPoint(17, 0, 0, 0, 15, True, ()),
            ],
        ),
        # Templated case (with consuming whitespace)
        (
            "{% for item in [1, 2] -%}\n"
            "SELECT *\n"
            "FROM some_table\n"
            "{{ 'UNION ALL\n' if not loop.last }}\n"
            "{%- endfor %}",
            "jinja",
            [
                # No initial indent (this is the first newline).
                # Importantly this first point - IS a newline
                # even though that newline segment is consumed
                # it should still be True here.
                _IndentPoint(1, 1, 0, 0, None, True, ()),
                # point between SELECT & *
                _IndentPoint(3, 1, 0, 1, 1, False, ()),
                # point after *
                _IndentPoint(5, -1, -1, 2, 1, True, (2,)),
                # point after FROM
                _IndentPoint(7, 1, 0, 1, 5, False, ()),
                # point after some_table
                _IndentPoint(9, -1, -1, 2, 5, True, (2,)),
                # point after ALL (we dedent down to the loop marker).
                _IndentPoint(13, -1, -1, 1, 9, True, ()),
                # There should be a loop marker here.
                # point after loop marker and before SELECT
                # (we indent back up after the loop).
                _IndentPoint(15, 1, 0, 0, 13, True, ()),
                # point between SELECT & *
                _IndentPoint(17, 1, 0, 1, 15, False, ()),
                # point after *
                _IndentPoint(19, -1, -1, 2, 15, True, (2,)),
                # point after FROM
                _IndentPoint(21, 1, 0, 1, 19, False, ()),
                # point after some_table (and before unused placeholder)
                _IndentPoint(23, -1, -1, 2, 19, True, (2,)),
                # Point after placeholder and dedenting down to endfor
                _IndentPoint(25, -1, -1, 1, 23, True, ()),
                # Point between endfor and end-of-file
                _IndentPoint(27, 0, 0, 0, 25, False, ()),
            ],
        ),
        # Templated case (with templated newline and indent)
        (
            "SELECT\n  {{'1 \n, 2'}}\nFROM foo",
            "jinja",
            [
                # After SELECT
                _IndentPoint(1, 1, 0, 0, None, True, ()),
                # NOTE: The newline inside the tag isn't reported.
                # After the templated section (hence why 7)
                _IndentPoint(7, -1, -1, 1, 1, True, ()),
                # After FROM
                _IndentPoint(9, 1, 0, 0, 7, False, ()),
                # After foo
                _IndentPoint(11, -1, -1, 1, 7, False, (1,)),
            ],
        ),
        # Templated case (with special marker slice that has no source string)
        (
            # The invisible special marker slice will be inserted immediately after
            # the first normal template section.
            "{% if True %}\n    SELECT 1;\n{% endif %}\n",
            "special_marker_inserter",
            [
                # No point at the start.
                # After the {% if True %} block: this should not yet indent because
                # there's still the upcoming zero-length special_marker
                # TemplateSegment.  This is handled similar to the trailing comment
                # test case.
                _IndentPoint(1, 0, 0, 0, None, False, ()),
                # After the zero-length special_marker TemplateSegment inserted by
                # the special templater: only after this do we want to indent.
                _IndentPoint(3, 1, 0, 0, None, True, ()),
                # After SELECT
                _IndentPoint(5, 1, 0, 1, 3, False, ()),
                # After 1
                _IndentPoint(7, -1, -1, 2, 3, False, (2,)),
                # After ;
                _IndentPoint(9, -1, -1, 1, 3, True, ()),
                # After {% endif %}
                _IndentPoint(11, 0, 0, 0, 9, True, ()),
            ],
        ),
    ],
)
def test_reflow__crawl_indent_points(raw_sql_in, templater, points_out, caplog):
    """Test _crawl_indent_points directly."""
    # Register the mock templater in this module.
    purge_plugin_manager()
    get_plugin_manager().register(sys.modules[__name__], name="reindent_test")

    config = FluffConfig(overrides={"dialect": "ansi", "templater": templater})
    root = parse_ansi_string(raw_sql_in, config)
    print(root.stringify())
    seq = ReflowSequence.from_root(root, config=config)
    with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules.reflow"):
        points = list(_crawl_indent_points(seq.elements))
    assert points == points_out


@pytest.mark.parametrize(
    "raw_sql_in,raw_sql_out",
    [
        # Trivial
        (
            "select 1",
            "select 1",
        ),
        # Initial Indent
        (
            "      select 1",
            "select 1",
        ),
        # Trailing Newline
        (
            "      select 1\n",
            "select 1\n",
        ),
        # Basic Multiline
        (
            "select\n1",
            "select\n  1",
        ),
        # Advanced Multiline
        (
            "select\n1+(\n2+3\n),\n4\nfrom foo",
            "select\n  1+(\n    2+3\n  ),\n  4\nfrom foo",
        ),
        (
            "select\n    1+(\n    2+3\n    ),\n    4\n    from foo",
            "select\n  1+(\n    2+3\n  ),\n  4\nfrom foo",
        ),
        # Multiple untaken indents. We should only indent as many
        # times as required.
        (
            "   select ((((\n1\n))))",
            "select ((((\n  1\n))))",
        ),
        (
            "select (((\n((\n3\n))\n)))",
            "select (((\n  ((\n    3\n  ))\n)))",
        ),
        # ### Templated Multiline Cases ###
        # NOTE: the templated tags won't show here, but they
        # should still be indented.
        # Trailing tag. NOTE: Last tag indented
        (
            "select\n1\n{% if true %}\n+ 2\n{% endif %}",
            "select\n  1\n  \n    + 2\n  ",
        ),
        # Cutting across the parse tree
        (
            "select\n1\n{% if true %}\n,2\nFROM a\n{% endif %}",
            # This set of template tags cuts across the parse
            # tree. We should indent them appropriately. In this case
            # that should mean "case 3", picking the lowest of the
            # existing indents which should mean no indent for either.
            # We also shouldn't indent the contents between them either
            # when taking this option.
            "select\n  1\n\n  ,2\nFROM a\n",
        ),
        # Template tags at file ends
        (
            "{% if true %}\nSELECT 1\n{% endif %}",
            "\n  SELECT 1\n",
        ),
        # Template loops:
        (
            "select\n  0,\n  {% for i in [1, 2, 3] %}\n    {{i}},\n  {% endfor %}\n  4",
            "select\n  0,\n  \n    1,\n  \n    2,\n  \n    3,\n  \n  4",
        ),
        # Correction and handling of hanging indents
        (
            "select 1, 2",
            "select 1, 2",
        ),
        (
            "select 1,\n2",
            "select\n  1,\n  2",
        ),
        (
            "select 1,\n       2",
            "select\n  1,\n  2",
        ),
        # A hanging example where we're modifying a currently empty point.
        (
            "select greatest(1,\n2)",
            "select greatest(\n  1,\n  2\n)",
        ),
        # Test handling of many blank lines.
        # NOTE:
        #    1. Initial whitespace should remain, because it's not an indent.
        #    2. Blank lines should also remain, because they're also not an indent.
        (
            "\n\n  \n\nselect\n\n\n\n    \n\n     1\n\n       \n\n",
            "\n\n  \n\nselect\n\n\n\n    \n\n  1\n\n       \n\n",
        ),
        # Templated cases.
        # NOTE: We're just rendering the fixed file in the templated space
        # so that for these tests we don't touch the fix routines. That's
        # why the template tags aren't visible - BUT THEIR INDENTS SHOULD BE.
        # This one is useful for ensuring the tags have the same indent.
        # ... first with a FROM
        (
            "SELECT\n"
            "    {{ 'a' }}\n"
            "    {% for c in ['d', 'e'] %}\n"
            "    ,{{ c }}_val\n"
            "    {% endfor %}\n"
            "FROM foo",
            "SELECT\n"
            "  a\n"
            "  \n"
            "    ,d_val\n"
            "  \n"
            "    ,e_val\n"
            "  \n"
            "FROM foo",
        ),
        # ... then without a FROM
        (
            "SELECT\n"
            "    {{ 'a' }}\n"
            "    {% for c in ['d', 'e'] %}\n"
            "    ,{{ c }}_val\n"
            "    {% endfor %}\n",
            "SELECT\n  a\n  \n    ,d_val\n  \n    ,e_val\n  \n",
        ),
        # This one is useful for if statements get handled right.
        # NOTE: There's a template loop in the middle.
        (
            "SELECT\n"
            "  {{ 'a' }}\n"
            "  {% for c in ['d', 'e'] %}\n"
            " {% if c == 'd' %}\n"
            "  ,{{ c }}_val_a\n"
            "    {% else %}\n"
            "  ,{{ c }}_val_b\n"
            "{% endif %}\n"
            "  {% endfor %}\n",
            "SELECT\n"
            "  a\n"
            "  \n"
            "    \n"
            "      ,d_val_a\n"
            "    \n"
            "  \n"
            "    \n"
            "      ,e_val_b\n"
            "    \n"
            "  \n",
        ),
        # Test leading templated newlines.
        # https://github.com/sqlfluff/sqlfluff/issues/4485
        (
            "{{ '\\n   \\n   ' }}\nSELECT 1",
            # NOTE: This looks a little strange, but what's important
            # here is that it doesn't raise an exception.
            "\n   \n   \nSELECT 1",
        ),
    ],
)
def test_reflow__lint_indent_points(raw_sql_in, raw_sql_out, default_config, caplog):
    """Test the lint_indent_points() method directly.

    Rather than testing directly, for brevity we check
    the raw output it produces. This results in a more
    compact test.
    """
    root = parse_ansi_string(raw_sql_in, default_config)
    print(root.stringify())
    seq = ReflowSequence.from_root(root, config=default_config)

    with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules.reflow"):
        elements, results = lint_indent_points(seq.elements, single_indent="  ")

    result_raw = "".join(elem.raw for elem in elements)
    assert result_raw == raw_sql_out, "Raw Element Check Failed!"

    # Now we've checked the elements - check that applying the fixes gets us to
    # the same place.
    print("Results:", results)
    anchor_info = compute_anchor_edit_info(fixes_from_results(results))
    fixed_tree, _, _, valid = apply_fixes(
        root, default_config.get("dialect_obj"), "TEST", anchor_info
    )
    assert valid, f"Reparse check failed: {fixed_tree.raw!r}"
    assert fixed_tree.raw == raw_sql_out, "Element check passed - but fix check failed!"


@pytest.mark.parametrize(
    "indent_line, forced_indents, expected_units",
    [
        # Trivial case of a first line.
        (
            _IndentLine(0, [_IndentPoint(0, 0, 0, 0, None, False, ())]),
            [],
            0,
        ),
        # Simple cases of a normal lines.
        (
            _IndentLine(3, [_IndentPoint(6, 0, 0, 3, 1, True, ())]),
            [],
            3,
        ),
        (
            # NOTE: Initial indent for *line* is different to *point*.
            # The *line* takes precedence.
            _IndentLine(1, [_IndentPoint(6, 0, 0, 3, 1, True, ())]),
            [],
            1,
        ),
        # Indents and dedents on the line break.
        # NOTE: The line indent still takes precedence here.
        (
            _IndentLine(3, [_IndentPoint(6, 1, 0, 3, 1, True, ())]),
            [],
            3,
        ),
        (
            _IndentLine(3, [_IndentPoint(6, -1, -1, 3, 1, True, ())]),
            [],
            3,
        ),
        # Handle untaken indents.
        (
            _IndentLine(3, [_IndentPoint(6, 0, 0, 3, 1, True, (1,))]),
            [],
            2,
        ),
        (
            _IndentLine(3, [_IndentPoint(6, 0, 0, 3, 1, True, (1, 2))]),
            [],
            1,
        ),
        (
            _IndentLine(3, [_IndentPoint(6, 0, 0, 3, 1, True, (2,))]),
            # Forced indent takes us back up.
            [2],
            3,
        ),
        (
            _IndentLine(3, [_IndentPoint(6, 0, 0, 3, 1, True, (3,))]),
            [],
            2,
        ),
        (
            _IndentLine(3, [_IndentPoint(6, 0, -1, 3, 1, True, (3,))]),
            # Untaken indent is pruned by trough.
            [],
            3,
        ),
    ],
)
def test_reflow__desired_indent_units(indent_line, forced_indents, expected_units):
    """Test _IndentLine.desired_indent_units() directly."""
    assert indent_line.desired_indent_units(forced_indents) == expected_units
